#include "xpttyd.h"

int handle_signals();
int old_version;
struct timeval tvp;
struct timezone tzp;
connection_table connections, *connection_ptr;
char buf[BUFSIZE];
extern int errno;

main(argc, argv)
int argc;
char *argv[];
{
	int loops, nfound;
	struct timeval timeout;
	fd_set efds, rfds, wfds;

	old_version = 0;
		/* check syntax for two args only */
	if ( (argc != 2) && (argc != 3) ) {
	    fprintf(stderr,"usage: xpttyd xpttydtab_file [log_file]\n");
	    exit(0);
	}	
		/* open log file and attach to stderr */
	if (argc == 3)
	    open_log_file(argv[2]);
	else
	    open_log_file(LOGFILE);
		/* parse the connection table passed on the command line,
		   open sockets, ptys and prep for connection to xterm */
	get_connection_table(argv[1]);
		/* at this point the ptys have been opened, local sockets
		   have been created, the X terms ip_addr and port numbers
		   have been assigned to a sockaddr_in.  We are basically
		   ready to establish the connection to the terminals ttys.
		   We'll now attempt to connect the ptys to the sockets. */
	fprintf(stderr,"AFter get_connection table...\n");
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP,  handle_signals);
	signal(SIGINT, handle_signals);
	signal(SIGQUIT, handle_signals);
	signal(SIGTERM, handle_signals);
		/* Set the conection_ptr to first table entry.  This ptr
		   will used throughout the program */
	connection_ptr = &connections;
	while (connection_ptr->next != NULL) {
	    connect_it();
	    connection_ptr = connection_ptr->next;
	}

		/* The plan now is to add all the pty and socket fds to a
		   select function and wait for someone to attempt a write.
		   We'll select for ready to read, or exception, do some IO
		   and loop.  We'll only read BUFSIZE chars on each IO to
		   prevent one connection from pigging out the others.  */

			/* forever loop */
	loops = 1;
	while (loops) {
			/* get the time so I check the timeout values for those
			   refused and timed out connections.  Basically,
			   check for retry time */
	 gettimeofday(&tvp,&tzp);
	  timeout.tv_sec = 15;
	  timeout.tv_usec = 0;
	  connection_ptr = &connections;
			/* zero the read, write and exception masks */
	  FD_ZERO (&rfds);
	  FD_ZERO (&wfds);
	  FD_ZERO (&efds);

			/* Loop through all the entries.  If
			   connection_ptr->next is NULL, the connection_ptr
			   is pointing at the last connection entry which is
			   unused.  (Strange but true) */
	  while (connection_ptr->next != NULL) {
			/* if we get a socket error, handle it.  The error
			   handler will reset the connection_ptr->error back
			   to 0 before return */
	    if (connection_ptr->error != 0 ) {
		handle_error();
	    }
			/* If this connection is in a holding pattern, see if
			   the timeout time has expired.  If it has, set
			   timeout to 0 so it can be added to the select. */
	    if (connection_ptr->timeout != 0 ) {
		if (connection_ptr->timeout < tvp.tv_sec ) {
			connection_ptr->timeout = 0;
	    		connect_it();
		}
	    }
			/* if the open pty and socket succeeded, add to the
			   select.  We add both to the select so we can listen
			   for data in either direction.  */
	    if ( (connection_ptr->ptyfd > 0) &&
	         (connection_ptr->local_socket > 0) &&
	         (connection_ptr->timeout == 0 ) ) {
	        FD_SET (connection_ptr->ptyfd, &rfds);
	        FD_SET (connection_ptr->local_socket, &rfds);
	        FD_SET (connection_ptr->ptyfd, &efds);
	        FD_SET (connection_ptr->local_socket, &efds);
	    }
	    if ( (connection_ptr->cbuf_cnt > 0) &&
	         (connection_ptr->timeout == 0 ) ) {
		timeout.tv_sec = 0;
		timeout.tv_usec = 50000;
	    }
	    connection_ptr = connection_ptr->next;
	  }
			/* Block until someone wants to talk  or timeout
			   occurs.  The timeout value set will be the worse
			   resolution of the timers (timeout and refused)
			   above. If we have unwritten data, set the value
			   lower. */
	  nfound = select(FD_SETSIZE, &rfds, &wfds, &efds, &timeout);
	  if (nfound == -1) {
	      if (errno != EINTR) {
	          perror("select error");
	          exit(0);
	      }
	  }
	  else if (nfound == 0) {  /* timeout */
	      connection_ptr = &connections;
				/* test for unwritten data */
	      while (connection_ptr->next != NULL) {
		if (connection_ptr->cbuf_cnt > 0) {
				/* try to write it */
		    do_com();
		}
	        connection_ptr = connection_ptr->next;
	      }
	  }
	  else {  /* not a timeout */
			/* first check for exceptions */
	      connection_ptr = &connections;
	      while (connection_ptr->next != NULL) {
	       if (connection_ptr->ptyfd > 0) {
	        if (FD_ISSET(connection_ptr->ptyfd, &efds)) {
	    	    FD_CLR(connection_ptr->ptyfd, &rfds);
			/* What can cause an exception upon a pty? */
		    exception_handle(connection_ptr->ptyfd);
	        }
	        if (FD_ISSET(connection_ptr->local_socket, &efds)) {
		    FD_CLR(connection_ptr->local_socket, &rfds);
		    exception_handle(connection_ptr->local_socket);
	        }
	       }
	       connection_ptr = connection_ptr->next;
	      }
			/* next check who can read */
	      connection_ptr = &connections;
	      while (connection_ptr->next != NULL) {
			/* if we still have unwritten data, try to write it
			   instead of reading anything new */
		if (connection_ptr->cbuf_cnt > 0) {
		    do_com();
		}
		else {		/* read new data */
	         if (connection_ptr->ptyfd > 0) {
	          if (FD_ISSET(connection_ptr->ptyfd, &rfds)) {
		    connection_ptr->tofd = connection_ptr->local_socket;
		    connection_ptr->fromfd = connection_ptr->ptyfd;
		    do_com();
		  }
				/* if empty in last direction, read the other */
		  if (connection_ptr->cbuf_cnt <= 0) {
	            if (FD_ISSET(connection_ptr->local_socket, &rfds)) {
		      connection_ptr->fromfd = connection_ptr->local_socket;
		      connection_ptr->tofd = connection_ptr->ptyfd;
		      do_com();
		    }
		  }
		} 
	       }
	       connection_ptr = connection_ptr->next;
	      }
	  }
        }
}

do_com()
{
	int nchars, wrote;
	int s_error_len, s_error;

			/* if error is set, handle it */
	if (connection_ptr->error != 0)
	    handle_error();
			/* if the connection is inactive, connect it */
	if (connection_ptr->active == 0) {
	    connect_it();
			/* if still inactive, return */
	    if (connection_ptr->active == 0)
		return;
	}
			/* first be sure we are able to write to the 
			   destination before reading from the source.  If we
			   don't, we will read unwritable data (not good) */
	if (able_to_write(connection_ptr->tofd, 250) == 0)
		return;
			/* check for socket error, just in case */
	s_error = 0;
	s_error_len = sizeof(s_error);
	getsockopt(connection_ptr->local_socket, SOL_SOCKET,
					SO_ERROR, &s_error, &s_error_len);
	if (s_error) {
		connection_ptr->error = s_error;
		errno = s_error;
		perror(NULL);
		return;
	}
			/* single buffer on the read, to allow others
			   a chance */
			/* test for unwritten data in cbuf.  If true, skip
			   the read and attempt to write the old buffer. */
	if (connection_ptr->cbuf_cnt <= 0) {
	    if ((connection_ptr->cbuf_cnt = 
			read(connection_ptr->fromfd,
			connection_ptr->cbuf, BUFSIZE)) == -1) {
		if (errno == EWOULDBLOCK) { /* no problem */
			connection_ptr->cbuf_cnt = 0; 
			return;
		}
						/* barf */
		perror("Comm read error, closing connection");
		fprintf(stderr,"Giving up on %s\n", connection_ptr->node_name);
		close_connection();
		return;
	    }
	    else if (connection_ptr->cbuf_cnt == 0) { /* eof reached */
		    exception_handle(connection_ptr->fromfd);
	    }
	    connection_ptr->written = 0;
	}
			/* Loop until data is written, error, or blocked */
	while (connection_ptr->cbuf_cnt > 0) {
	    if ((wrote = write(connection_ptr->tofd,
			&(connection_ptr->cbuf[connection_ptr->written]),
					connection_ptr->cbuf_cnt)) == -1) {
		if (errno == EWOULDBLOCK) { /* spin til done */
			wrote = 0;
			return;
		}
		else if (errno == EPIPE) {    /* no problem */
		    fprintf(stderr,"pipe error: %s, lost %d bytes of data\n",
			connection_ptr->node_name,
			connection_ptr->cbuf_cnt - connection_ptr->written); 
		    connection_ptr->cbuf_cnt = 0;
		    connection_ptr->written = 0;
		    perror(NULL);
		    exception_handle(connection_ptr->tofd);
		    return;
		}
		else {			/* barf */
		    perror("Comm write error, closing connection");
		    connection_ptr->cbuf_cnt = 0;
		    connection_ptr->written = 0;
		    fprintf(stderr,"Giving up on %s\n",
					connection_ptr->node_name);
		    close_connection();
		}
	    }
	    else if (wrote == 0) {	/* eof */
		exception_handle(connection_ptr->tofd);
		connection_ptr->cbuf_cnt = 0;
		connection_ptr->written = 0;
		return;
	    }
	    connection_ptr->cbuf_cnt -= wrote;
	    connection_ptr->written += wrote;
	}
}

exception_handle(fd)
int fd;
{
				/* close the ports and reopen them */
	if (fd == connection_ptr->ptyfd) {
	    if (connection_ptr->ptyfd > 0) {
		close(connection_ptr->ptyfd);
	    }
    	    connection_ptr->active = 0;
	    if (old_version)
	      connection_ptr->ptyfd = old_open_pseudo_port(connection_ptr->pty);
	    else
	      connection_ptr->ptyfd = new_open_pseudo_port(connection_ptr->pty);
	}
	else if (fd == connection_ptr->local_socket) {
	    if (connection_ptr->local_socket > 0) {
		close(connection_ptr->local_socket);
		set_linger(connection_ptr->local_socket, 0);
	    }
    	    connection_ptr->active = 0;
	    connection_ptr->local_socket = get_local_socket();
	}
}

get_connection_table(table_file)
char *table_file;
{
	FILE *fp, *fopen();
	int status;
	char *p1;
	
	fprintf(stderr,"get_connection_table.....\n");

	if ( (fp = fopen(table_file, "r")) == NULL) {
		fprintf(stderr,"Error opening xpttyd table:%s\n",table_file);
		perror("get_connection_table");
		exit(errno);
	}
			/* set connection_ptr to first structure */
	connection_ptr = &connections;
			/* set the structure to all zeros */
	bzero(connections, sizeof(connection_table));
	status = 0;
	while (status != EOF) {
		/* some fscanf's don't work very well, so we'll get a line
		   and test for comment or blank line before conversion */
		/* get a line */
	  if ((p1 = fgets(buf, BUFSIZE, fp)) == NULL) {
	    status = EOF;
	  }
	  else {
	    status = 0;
	    p1 = &buf[0];
			/* remove white space */
	    while (isspace((int)(*p1))) {
	        p1++;
	    }

			/* if first char is not #, newline, or null, use it */
	    if ( (*p1 != '#') && (*p1 != '\n') && (*p1 != '\0'))
	        status = sscanf(p1, "%s %d %s", &connection_ptr->node_name[0],
					        &connection_ptr->port,
					        &connection_ptr->pty[0]);
			/* if conversion is correct, status will be 3 */
	    if (status == 3) {
			/* open local socket for communication to xterm */
	        connection_ptr->local_socket = get_local_socket();
		fprintf(stderr,"LOG: connect %s to %s port %d\n",
						connection_ptr->pty,
						connection_ptr->node_name,
						connection_ptr->port);
			/* set up xterm sockadd_in */
	        get_xterm_sockaddr(connection_ptr->node_name,
				   connection_ptr->port,
				   &connection_ptr->xterm);
			/* open the pseudo tty port */
	        if (old_version)
			{
			  fprintf(stderr,"OLD VERSION\n");
	          connection_ptr->ptyfd =
				old_open_pseudo_port(connection_ptr->pty);
			}
	        else
			{
			  fprintf(stderr,"NEW VERSION\n");
	          connection_ptr->ptyfd = 
				new_open_pseudo_port(connection_ptr->pty);
             }
print_connection_table(connection_ptr);

			/* alloc space for next connection */
		if ((connection_ptr->next =
		    (connection_table *) malloc(sizeof(connection_table)) ) ==
		    NULL) {
			perror("get_connection_table malloc failed");
			exit(0);
		}
			/* get ready for next connection */
		connection_ptr = connection_ptr->next;
		bzero(connection_ptr, sizeof(connection_table));
	    }
	  }
	}
}

int
get_xterm_sockaddr(node_name, port, xterm)
char node_name[];
int port;
struct sockaddr_in *xterm;
{
	struct hostent *hp, *gethostbyname();
	char node[20];
	long temp;

	fprintf(stderr,"In get_xterm_sockaddr .... \n");

			/* is the node name or number */
	if ( !(isdigit((int)(node_name[0]))) ) {
		if ( ((hp = gethostbyname(node_name)) == NULL) ||
		      (hp->h_addrtype != AF_INET) ) {
			fprintf(stderr,"%s: unknown node\n",node_name);
			perror("get_xterm_sockaddr");
			return(NULL);
		}
		bcopy(hp->h_addr, &xterm->sin_addr.s_addr, hp->h_length);
	}
	else {
		strcpy(node,node_name);
		temp = atoi(strtok(node,"."));
		xterm->sin_addr.s_addr = temp << 24;
		temp = atoi(strtok(NULL,"."));
		xterm->sin_addr.s_addr |= temp << 16;
		temp = atoi(strtok(NULL,"."));
		xterm->sin_addr.s_addr |= temp << 8;
		temp = atoi(strtok(NULL,"."));
		xterm->sin_addr.s_addr |= temp;
		xterm->sin_addr.s_addr = htonl(xterm->sin_addr.s_addr);
	}
				/* be sure order is correct */
	xterm->sin_family = AF_INET;
	xterm->sin_port = htons(port);
	fprintf(stderr,"Exiting  get_xterm_sockaddr .... \n");
	return(1);
}

int
get_local_socket()
{
	int sock;
					/* open socket on host */
	if ( (sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("Error opening local socket");
		return(0);
	}
	set_linger(sock, 1);
			/* set FNDELAY so we don't ever block */
	if (fcntl(sock, F_SETFL, FNDELAY)) {
		perror("Error setting FNDELAY");
	}

	fprintf(stderr,"get_local_socket: socket is %d\n",sock);
	return(sock);
}

set_linger(sock, state)
int sock, state;
{
#ifndef hpux
	struct linger net_linger;

        net_linger.l_onoff = state;   /* enable linger on close of socket */
        net_linger.l_linger = 360; /* 360 second timeout */
#else  /* is hpux */
	int net_linger;

        net_linger = state;
#endif /*hpux*/

        setsockopt(sock, SOL_SOCKET, SO_LINGER, &net_linger,
							 sizeof(net_linger));
}

int
connect_pty_xptty(cptr)
connection_table *cptr;
{
	int stat;
	
	fprintf(stderr,"Entering connect_pty_xptty....\n");
		/* test to be sure we have a local socket, and xterm port */
	if ( (cptr->local_socket == 0) || (cptr->xterm.sin_port == 0) ) {
		fprintf(stderr,"No connection bad socket %d %d\n",
				cptr->local_socket, cptr->xterm.sin_port);
		return(0);
	}
		/* connect em.  We need to work on the connection time out.
		   An unreachable terminal can cost us 60 seconds. */
	fprintf(stderr,"connect_pty_xptty: local socket is %d\n",cptr->local_socket);
	stat = connect(cptr->local_socket, &cptr->xterm, sizeof(cptr->xterm));
	connection_ptr->pending++;
	if (stat == -1) {
	     if ( (errno != EINPROGRESS) &&
	          (errno != EISCONN) ) {
	       	fprintf(stderr,"Error connecting to %s socket %d errno %d",
				cptr->node_name, cptr->xterm.sin_port, errno);
	       	perror(" ");
	     }
	    return(0);
	}
	return(1);
}

int
old_open_pseudo_port(pseudo_port)
char *pseudo_port;
{
	int fd;
	char msg[80];

	if ( (fd = open (pseudo_port, O_RDWR, NULL)) == -1 ) {
		sprintf(msg,"Error opening pseudo_port %s\n",pseudo_port);
		perror(msg);
	}
	else {			/* set not blocking */
	    if (fcntl(fd, F_SETFL, FNDELAY)) {
		perror("Error setting FNDELAY");
	    }
	}
	return(fd);
}

int
new_open_pseudo_port(pseudo_port)
char *pseudo_port;
{
	int fd, cnt;
	char msg[80];

		/* This routine calls get_pty which returns an open fd.  If we
		   have an open fd, link it to the pseudo_port */
	if ((get_pty(&fd)) != 0) {
		perror("Can't find available pty port");
		return(-1);
	}
	else {			/* set not blocking */
	    if (fcntl(fd, F_SETFL, FNDELAY)) {
		perror("Error setting FNDELAY");
	    }
	}
		/* if the oldtty is the same as the newtty skip the 
		   unlink and relink stage.  This should solve the problem
		   of looping scripts outputing little data and creating
		   a file between my unlink/link stage. The possibility
		   still exists, if I end up with a different tty then I
		   had the first time.  */
	if (strcmp(connection_ptr->oldttydev, connection_ptr->ttydev)) {
	    strcpy(connection_ptr->oldttydev, connection_ptr->ttydev);
			/* unlink the existing pseudo port */
	    cnt = 0;
	    while (cnt < LINK_RETRY) {
	      (void) unlink(pseudo_port);
			/* link ttydev to the pseudo port */
	      if ((link(connection_ptr->ttydev, pseudo_port)) == -1) {
		if (errno == EEXIST) { /* try again */
		    cnt++;
		    if (cnt >= LINK_RETRY) {
		        perror("Link");
		        return(-1);
		    }
		}
		else {
		    perror("Link");
		    return(-1);
		}
	      }
	      else
	          break;
	    }
	    if ((chown(connection_ptr->ttydev, 0, 0)) == -1) {
		perror("Chown");
	    }
	    if ((chmod(connection_ptr->ttydev, 0666)) == -1) {
		perror("Chmod");
	    }
	}
	return(fd);
}

print_connection_table(tab)
connection_table *tab;
{
	printf("active = %d ptyfd = %d local_socket = 0x%x ",
		 tab->active, tab->ptyfd, tab->local_socket);
	printf("xterm addr 0x%x xterm port = %d\n", 
		tab->xterm.sin_addr.s_addr, tab->xterm.sin_port);
}

int able_to_write(fd, usecs)
int fd, usecs;
{
    struct timeval timeout;
    int stat;
    fd_set efds, rfds, wfds;

    /* establish timeout values */
    timeout.tv_sec = 0;
    timeout.tv_usec = usecs;

    FD_ZERO (&wfds);
    FD_ZERO (&efds);
    FD_SET (fd, &wfds);
    FD_SET (fd, &efds);
    if((stat = select(FD_SETSIZE, (int *)0, &wfds, &efds, &timeout)) < 0) {
        fprintf(stderr,"able_to_write select returned %d, aborting\n", stat);
	return(0);
    }
    return (FD_ISSET (fd, &wfds));
}

read_conn(tab)
connection_table *tab;
{
	int nchars;

	while ((nchars = read(tab->ptyfd, buf, BUFSIZE)) > 0) {
		buf[nchars] = '\0';
		printf("read %d %s",nchars,buf);
	}
}

open_log_file(fname)
char *fname;
{
	int fd;
		/* open log file for errors and messages */
	if ((fd = open(fname, O_RDWR | O_TRUNC | O_CREAT, 0644)) == -1) {
	    fprintf(stderr,"Can't open logfile '%s' using stderr\n",fname);
	}
	else {
	    close(2); /* close stderr, and connect it to the log file */
	    if (dup2(fd, 2) != 2) {
		perror("dup:");
		exit(0);
	   }
	}
}

connect_it()
{
	int stat, s_error_len, s_error;

	fprintf(stderr,"Entering connect it .... \n");

			/* make a single attempt to connect - non blocking */
	if (connection_ptr->pending == 0) {
	      connection_ptr->active = connect_pty_xptty(connection_ptr);
	}
			/* connection should, fail due to non-blocking IO.
			   An error of EINPROGRESS is typical.  The
			   able_to_write function will return 0 until the
			   connection succeeds or fails.  The connection
			   should then change state to active if the connection
			   succeeded.  If a failure occurred, store the error
			   code and proceed accordingly */
	stat = able_to_write(connection_ptr->local_socket, 1000);
			/* if we can't connect, increment pending & return */
	if (stat != 0) {
	    s_error = 0;
	    s_error_len = sizeof(s_error);
	    getsockopt(connection_ptr->local_socket, SOL_SOCKET,
					SO_ERROR, &s_error, &s_error_len);
 	    connection_ptr->error = s_error;
	    if (s_error) {
	        connection_ptr->active = 0;
 	        connection_ptr->pending = 0;
	        errno = s_error;
 	        if (!(connection_ptr->logged & 1)) {
		    fprintf(stderr, "Connection to %s port %d failed: ",
			connection_ptr->node_name, connection_ptr->port);
 	            connection_ptr->logged = 1;
	            perror(NULL);
		}
	      if (connection_ptr->local_socket > 0) {
		close(connection_ptr->local_socket);
		set_linger(connection_ptr->local_socket, 0);
	      }
	      connection_ptr->local_socket = get_local_socket();
	    }
	    else {
 	        if (!(connection_ptr->logged & 2)) {
		   fprintf(stderr, "Connected to %s port %d\n",
			connection_ptr->node_name, connection_ptr->port);
		}
 	        connection_ptr->logged = 2;
	        connection_ptr->active = 1;
 	        connection_ptr->pending = 0;
	    }
	}

	fprintf(stderr,"Exiting connect_it....\n");
}

close_connection()
{
    connection_ptr->active = 0;
    if (connection_ptr->local_socket > 0) {
	set_linger(connection_ptr->local_socket, 0);
	close(connection_ptr->local_socket);
	connection_ptr->local_socket = 0;
    }
    if (connection_ptr->ptyfd > 0) {
	close(connection_ptr->ptyfd);
	connection_ptr->ptyfd = 0;
    }
}

handle_error()
{
	switch (connection_ptr->error) {
		case 0:
			break;
		case EINPROGRESS:
		case EALREADY:
				/* ignore these */
			break;
		case ETIMEDOUT:
			connection_ptr->timeout = tvp.tv_sec + 
						TIMEDOUT_RETRY_INTERVAL;
				/* xterm's off, ping would probably fail */
			break;
		case ECONNREFUSED:
			connection_ptr->timeout = tvp.tv_sec + 
						CONNREFUSED_RETRY_INTERVAL;
				/* xterm's alive, but doesn't want to speak */
			break;
		case ECONNRESET:
			connection_ptr->timeout = tvp.tv_sec + 
						CONNREFUSED_RETRY_INTERVAL;
				/* xterm's doesn't want to speak, unknown why */
 	                connection_ptr->logged = 0;
			exception_handle(connection_ptr->local_socket);
			break;
		default:
				/* fatal error, requires restart of daemon */
			fprintf(stderr,"Giving up on %s\n",
						connection_ptr->node_name);
			close_connection();
			break;

	}
	connection_ptr->error = 0;
}


handle_signals(sig, code, sc)
int sig, code;
struct sigcontext *sc;
{
	switch (sig) {
		case SIGHUP:	/* reread the tab file */
			break;
		case SIGINT:	/* interrupt signal - exit */
		case SIGQUIT:	/* quit signal - exit */
		case SIGTERM:	/* terminate signal - exit */
                        /* unlink the existing pseudo ports */
			connection_ptr = &connections;
			while (connection_ptr->next != NULL) {
#if defined(sgi) || (defined(umips) && defined (SYSTYPE_SYSV))
				(void) chmod(connection_ptr->pty, 0600);
#endif
			        (void) unlink(connection_ptr->pty);
				connection_ptr = connection_ptr->next;
			}
			exit(0);
	}
}

/* This function opens up a pty master and stuffs its value into pty.
 * If it finds one, it returns a value of 0.  If it does not find one,
 * it returns a value of !0.  This routine is designed to be re-entrant,
 * so that if a pty master is found and later, we find that the slave
 * has problems, we can re-enter this function and get another one.
 */

get_pty (pty)
int *pty;
{
#if defined(sgi) || (defined(umips) && defined (SYSTYPE_SYSV))
	struct stat fstat_buf;
#endif
	strcpy(connection_ptr->ttydev, TTYDEV);
	strcpy(connection_ptr->ptydev, PTYDEV);
#if defined(SYSV) && defined(SYSV386)
        /*
	  The order of this code is *important*.  On SYSV/386 we want to open
	  a /dev/ttyp? first if at all possible.  If none are available, then
	  we'll try to open a /dev/pts??? device.
	  
	  The reason for this is because /dev/ttyp? works correctly, where
	  as /dev/pts??? devices have a number of bugs, (won't update
	  screen correcly, will hang -- it more or less works, but you
	  really don't want to use it).
	  
	  Most importantly, for boxes of this nature, one of the major
	  "features" is that you can emulate a 8086 by spawning off a UNIX
	  program on 80386/80486 in v86 mode.  In other words, you can spawn
	  off multiple MS-DOS environments.  On ISC the program that does
	  this is named "vpix."  The catcher is that "vpix" will *not* work
	  with a /dev/pts??? device, will only work with a /dev/ttyp? device.
	  
	  Since we can open either a /dev/ttyp? or a /dev/pts??? device,
	  the flag "IsPts" is set here so that we know which type of
	  device we're dealing with in routine spawn().  That's the reason
	  for the "if (IsPts)" statement in spawn(); we have two different
	  device types which need to be handled differently.
	  */
        if (pty_search(pty) == 0)
	    return 0;
#endif /* SYSV && SYSV386 */
#ifdef ATT
	if ((*pty = open ("/dev/ptmx", O_RDWR)) < 0) {
	    return 1;
	}
#if defined(SVR4) || defined(SYSV386)
	strcpy(connection_ptr->ttydev, ptsname(*pty));
#if defined (SYSV) && defined(SYSV386)
	IsPts = True;
#endif
#endif
	return 0;
#else /* ATT else */
#ifdef __convex__
        {
	    char *pty_name, *getpty();

	    while ((pty_name = getpty()) != NULL) {
		if ((*pty = open (pty_name, O_RDWR)) >= 0) {
		    strcpy(connection_ptr->ptydev, pty_name);
		    strcpy(connection_ptr->ttydev, pty_name);
		    connection_ptr->ttydev[5] = 't';
		    return 0;
		}
	    }
	    return 1;
	}
#endif /* __convex__ */
#ifdef USE_GET_PSEUDOTTY
	return ((*pty = getpseudotty (&connection_ptr->ttydev,
					&connection_ptr->ptydev)) >= 0 ? 0 : 1);
#else
#if defined(sgi) || (defined(umips) && defined (SYSTYPE_SYSV))
/*
	struct stat fstat_buf;
*/

	*pty = open ("/dev/ptc", O_RDWR);
	if (*pty < 0 || (fstat (*pty, &fstat_buf)) < 0) {
	  return(1);
	}
	sprintf (connection_ptr->ttydev, "/dev/ttyq%d",
						minor(fstat_buf.st_rdev));
#ifndef sgi
	sprintf (connection_ptr->ptydev, "/dev/ptyq%d",
						minor(fstat_buf.st_rdev));
/*
	if ((*tty = open (connection_ptr->ttydev, O_RDWR)) < 0) {
	  close (*pty);
	  return(1);
	}
*/
#endif /* !sgi */
	/* got one! */
	return(0);
#else /* sgi or umips */

	return pty_search(pty);

#endif /* sgi or umips else */
#endif /* USE_GET_PSEUDOTTY else */
#endif /* ATT else */
}

/*
 * Called from get_pty to iterate over likely pseudo terminals
 * we might allocate.  Used on those systems that do not have
 * a functional interface for allocating a pty.
 * Returns 0 if found a pty, 1 if fails.
 */
int pty_search(pty)
int *pty;
{
    static int devindex, letter = 0;

#ifdef CRAY
    for (; devindex < 256; devindex++) {
	sprintf (connection_ptr->ttydev, "/dev/ttyp%03d", devindex);
	sprintf (connection_ptr->ptydev, "/dev/pty/%03d", devindex);

	if ((*pty = open (connection_ptr->ptydev, O_RDWR)) >= 0) {
	    /* We need to set things up for our next entry
	     * into this function!
	     */
	    (void) devindex++;
	    return 0;
	}
    }
#else /* CRAY */
#ifdef SCO
    for (; devindex < 90; devindex++) {
	sprintf (connection_ptr->ttydev, "/dev/ttyp%02d", devindex);
	sprintf (connection_ptr->ptydev, "/dev/pty%02d", devindex);

	if ((*pty = open (connection_ptr->ptydev, O_RDWR)) >= 0) {
	    /* We need to set things up for our next entry
	     * into this function!
	     */
	    (void) devindex++;
	    return 0;
	}
    }
#endif /* SCO */
    while (PTYCHAR1[letter]) {
	connection_ptr->ttydev [strlen(connection_ptr->ttydev) - 2] =
		 connection_ptr->ptydev [strlen(connection_ptr->ptydev) - 2] =
	         PTYCHAR1 [letter];

	while (PTYCHAR2[devindex]) {
	    connection_ptr->ttydev [strlen(connection_ptr->ttydev) - 1] =
		connection_ptr->ptydev [strlen(connection_ptr->ptydev) - 1] =
		PTYCHAR2 [devindex];
	    if ((*pty = open (connection_ptr->ptydev, O_RDWR)) >= 0) {
		/* We need to set things up for our next entry
		 * into this function!
		 */
		(void) devindex++;
		return 0;
	    }
	    devindex++;
	}
	devindex = 0;
	(void) letter++;
    }
#endif /* CRAY else */
    /*
     * We were unable to allocate a pty master!  Return an error
     * condition and let our caller terminate cleanly.
     */
    return 1;
}

